from different_wrapping.utils import docker_port_string_get_external_port


def generate_ingress(service_name, service, challenge, args):
    # First we need to determine the port that the ingress is supposed to be pointed towards
    if "ports" not in service.container_dict:
        raise RuntimeError(
            f"Tried to create ingress for {service_name} but it has no ports exposed(and therefore no k8s service to point to)"
        )

    external_ports = [
        docker_port_string_get_external_port(port)
        for port in service.container_dict["ports"]
    ]

    if len(external_ports) == 0:
        raise RuntimeError(
            f"Unable to create ingress for {service_name} as there are no ports"
        )
    elif len(external_ports) > 1:
        raise RuntimeError(
            f"Unable to determine external port for {service_name} as there are more than one ports exposed"
        )

    # Determine DNS name
    host = service.get_dns_name(args.dns_host)

    return {
        "apiVersion": "networking.k8s.io/v1",
        "kind": "Ingress",
        "metadata": {
            "name": f"ingress-{challenge.name()}-{service_name}",
        },
        "spec": {
            "ingressClassName": "TODO",
            "rules": [
                {
                    "host": host,
                    "http": {
                        "paths": [
                            {
                                "path": "/",
                                "pathType": "Prefix",
                                "backend": {
                                    "service": {
                                        "name": service_name,
                                        "port": {"number": int(external_ports[0])},
                                    }
                                },
                            }
                        ]
                    },
                }
            ],
        },
    }



# Responsible for converting the docker-compose style manifests to a k8s kustomize project
import shutil
import subprocess
from pathlib import Path

from different_wrapping.ingress import generate_ingress
from different_wrapping.gateway import generate_gateway, generate_tcp_gateway

import yaml

import logging

logger = logging.getLogger(__name__)


def tempdir():
    """Debugging purposes"""
    import uuid

    directory = Path("/tmp") / str(uuid.uuid4())
    directory.mkdir(parents=True)
    return directory


def build_challenge_kustomize(files, out_file):
    """Builds the kustomize manifest for a challenge"""
    logger.info("Building for %s" % files)
    manifest = {
        "apiVersion": "kustomize.config.k8s.io/v1beta1",
        "kind": "Kustomization",
        "resources": [f.name for f in files],
    }

    with out_file.open("w") as f:
        yaml.dump(manifest, f)


def build_kustomize_manifest(challenges, folder):
    """Builds the topmost kustomize manifest which includes all challenges"""
    manifest = {
        "apiVersion": "kustomize.config.k8s.io/v1beta1",
        "kind": "Kustomization",
        "resources": [
            "./%s" % challenge.relative_path()
            for challenge in filter(
                lambda challenge: (
                    challenge.has_containers and challenge.has_static_challenges()
                ),
                challenges,
            )
        ],
    }

    out_file = folder / Path("kustomization.yml")

    with out_file.open("w") as f:
        yaml.dump(manifest, f)


def build_k8s_manifests(challenge, output, filter_lambda, args):
    """Builds k8s manifests to a given base directory(organized by challenge type).
    selector_lambda allows filtering of what containers to use"""
    logger.info("Processing %s" % challenge.relative_path())

    # with tempfile.TemporaryFile() as compose_file:
    # with tempfile.TemporaryDirectory() as workdir:
    workdir = Path(tempdir())

    service_dict = {
        key: value.container_dict
        for key, value in filter(filter_lambda, challenge.containers.items())
    }
    if len(service_dict.keys()) == 0:
        logger.info(
            "Challenge %s has no services valid for %s - skipping"
            % (challenge.relative_path(), str(output))
        )
        return

    compose_file = workdir / "docker-compose.yml"

    logger.info("Writing to %s" % compose_file)
    compose_data = {"version": "3", "services": service_dict}
    with compose_file.open("w") as f:
        yaml.dump(compose_data, f)

    kompose_args = ["kompose", "-f", "docker-compose.yml", "convert"]

    result = subprocess.run(kompose_args, cwd=workdir, capture_output=True)

    if result.returncode != 0:
        logger.error("Failed to run kompose! stderr: \n%s" % result.stderr)

    logger.info(
        "Kompose returned %d\nstdout %s\nstderr %s"
        % (result.returncode, result.stdout, result.stderr)
    )

    # Generate ingress
    generate_challenge_ingresses(challenge, workdir, args)

    # Manifest files to be copied
    files = list(
        filter(lambda file: "docker-compose" not in file.name, workdir.iterdir())
    )
    logger.info("Files: %s" % len(files))
    if not args.dry:
        # Create the output folder
        output_dir = output / Path(challenge.relative_path())
        output_dir.mkdir(parents=True, exist_ok=True)

        # Build kustomize file
        kustomize_file = output_dir / Path("kustomization.yml")
        build_challenge_kustomize(files, kustomize_file)

        # Copy over files
        for file in files:
            dest = output_dir / Path(file.name)
            logger.debug("%s -> %s" % (file, dest))
            shutil.copyfile(file, dest)

        logger.debug("Done copying files")


def generate_challenge_ingresses(challenge, outdir, args):
    outfile = outdir / "different-wrapping-ingress.yaml"

    ingresses = []

    for service_name, service in challenge.containers.items():
        logger.debug(service_name)
        if service.has_label("no.tghack.http"):
            logger.info("Container has http label, creating ingress")
            if args.ingress_type == "ingress":
                ingresses.append(
                    generate_ingress(service_name, service, challenge, args)
                )
            else:
                ingresses.append(
                    generate_gateway(service_name, service, challenge, args)
                )
        elif service.has_label("no.tghack.tcp"):
            logger.info("Container has TCP ingress, creating TCP route")
            if args.ingress_type == "ingress":
                raise RuntimeError(
                    "Ingress type ingress not supported with TCP challenges"
                )
            else:
                ingresses = ingresses + generate_tcp_gateway(
                    service_name, service, challenge, args
                )
        else:
            logger.info("Container has no ingress requirement")

    if len(ingresses) > 0:
        with outfile.open("w") as f:
            f.write("\n---\n".join([yaml.dump(ingress) for ingress in ingresses]))



def generate_gateway(service_name, service, challenge, args):
    # First we need to determine the port that the ingress is supposed to be pointed towards
    if "ports" not in service.container_dict:
        raise RuntimeError(
            f"Tried to create gateway for {service_name} but it has no ports exposed(and therefore no k8s service to point to)"
        )

    external_ports = service.ports()

    if len(external_ports) == 0:
        raise RuntimeError(
            f"Unable to create gateway for {service_name} as there are no ports"
        )
    elif len(external_ports) > 1:
        raise RuntimeError(
            f"Unable to determine external port for {service_name} as there are more than one ports exposed"
        )

    # Determine DNS name
    host = service.get_dns_name(args.dns_host)

    return {
        "apiVersion": "gateway.networking.k8s.io/v1",
        "kind": "HTTPRoute",
        "metadata": {
            "name": f"gateway-{challenge.safe_name()}-{service_name}",
        },
        "spec": {
            "parentRefs": [
                {
                    "name": args.gateway_parent_name,
                    "namespace": args.gateway_parent_namespace,
                    "sectionName": "https",
                    "kind": "Gateway",
                    "group": "gateway.networking.k8s.io",
                }
            ],
            "hostnames": [host],
            "rules": [
                {
                    "matches": [{"path": {"type": "PathPrefix", "value": "/"}}],
                    "backendRefs": [
                        {
                            "name": service_name,
                            "port": int(external_ports[0]),
                            "weight": 1,
                            "group": "",
                            "kind": "Service",
                        }
                    ],
                }
            ],
        },
    }


def generate_tcp_gateway(service_name, service, challenge, args):
    # First we need to determine the port that the ingress is supposed to be pointed towards
    if "ports" not in service.container_dict:
        raise RuntimeError(
            f"Tried to create TCP gateway for {service_name} but it has no ports exposed(and therefore no k8s service to point to)"
        )

    external_ports = service.ports()

    if len(external_ports) == 0:
        raise RuntimeError(
            f"Unable to create TCP gateway for {service_name} as there are no ports"
        )
    elif len(external_ports) > 1:
        raise RuntimeError(
            f"Unable to determine external port for {service_name} as there are more than one ports exposed"
        )

    # Determine DNS name
    host = service.get_dns_name(args.dns_host)

    if args.tcp_gateway_openstack_lb_id is None:
        raise RuntimeError("Unable to create TCP gateway - no LB id is set")

    gateway_name = f"tcp-gateway-{challenge.safe_name()}-{service_name}"

    gateway = {
        "apiVersion": "gateway.networking.k8s.io/v1",
        "kind": "Gateway",
        "metadata": {"name": gateway_name},
        "spec": {
            "gatewayClassName": "envoy-tcp-chall",
            "infrastructure": {
                "annotations": {
                    "external-dns.alpha.kubernetes.io/hostname": host,
                    "external-dns.alpha.kubernetes.io/ttl": "60",
                    "loadbalancer.openstack.org/enable-health-monitor": "true",
                    "loadbalancer.openstack.org/proxy-protocol": "false",
                    "loadbalancer.openstack.org/load-balancer-id": args.tcp_gateway_openstack_lb_id,
                }
            },
            "listeners": [
                {
                    "protocol": "TCP",
                    "port": int(external_ports[0]),
                    "name": "challenge",
                    "allowedRoutes": {
                        "namespaces": {"from": "Same"},
                        "kinds": [
                            {"kind": "TCPRoute", "group": "gateway.networking.k8s.io"}
                        ],
                    },
                }
            ],
        },
    }

    tcproute = {
        "apiVersion": "gateway.networking.k8s.io/v1alpha2",
        "kind": "TCPRoute",
        "metadata": {
            "name": f"tcproute-{challenge.safe_name()}-{service_name}",
        },
        "spec": {
            "parentRefs": [
                {
                    "name": gateway_name,
                    "sectionName": "challenge",
                    "group": "gateway.networking.k8s.io",
                    "kind": "Gateway",
                }
            ],
            "rules": [
                {
                    "backendRefs": [
                        {
                            "name": service_name,
                            "port": int(external_ports[0]),
                            "weight": 1,
                            "group": "",
                            "kind": "Service",
                        }
                    ]
                }
            ],
        },
    }

    return [gateway, tcproute]



from pathlib import Path

import yaml
import re

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from different_wrapping.compose import Container

import logging

logger = logging.getLogger(__name__)


class Challenge:
    def __init__(self, path: Path):
        self.path = path
        self.has_containers = True

        self.containers = {}
        logger.info("Loading challenge: %s" % self.relative_path())

        docker_compose = path / Path("docker-compose.yml")
        if docker_compose.exists():
            logger.debug("Docker-compose file found, parsing")
            self._read_docker_compose(docker_compose)
        else:
            compose = path / Path("compose.yml")
            if compose.exists():
                logger.debug("compose.yml found, parsing")
                self._read_compose(compose)
            else:
                self.has_containers = False

    def _read_docker_compose(self, compose_file):
        with compose_file.open() as f:
            compose = yaml.load(f, Loader)

        if compose["version"] != "3":
            raise RuntimeError(
                "Error parsing %s - only version 3 of compose is supported. See https://docs.docker.com/compose/compose-file/compose-file-v3/"
                % compose_file
            )

        for key, service in compose["services"].items():
            self.containers[key] = Container(key, self, service)

    def _read_compose(self, compose_file):
        with compose_file.open() as f:
            compose = yaml.load(f, Loader)

        for key, service in compose["services"].items():
            self.containers[key] = Container(key, self, service)

    def category(self):
        return self.path.parent.name

    def name(self):
        return self.path.name

    def cyber_name(self):
        # Cyberlandslaget hack
        with open(self.path / "challenge.yaml", "r") as f:
            try:
                parsed = yaml.safe_load(f)
            except Exception:
                raise RuntimeError(
                    f"Unable to parse challenge.yml for {self.path} - bad yaml"
                )

            if (
                "title" in parsed
                and "name" in parsed
                and parsed["title"] != parsed["name"]
            ):
                raise RuntimeError(
                    f'Unable to parse challenge.yml for {self.path} - both title and name field exist but they don\'t agree! "{parsed["title"]}" vs "{parsed["name"]}"'
                )

            if "title" in parsed:
                return parsed["title"]

            if "name" in parsed:
                return parsed["name"]

            raise RuntimeError(
                f"Unable to parse challenge.yml for {self.path} - neither title nor name field is provided."
            )

    def safe_name(self):
        return re.sub(r"[^a-zA-Z0-9]", "-", self.name()).lower()

    def relative_path(self):
        return "%s/%s" % (self.category(), self.name())

    def challenge_id(self):
        # Cyberlandslaget hack
        return ("%s/%s" % (self.category(), self.cyber_name())).lower()

    def determine_version(self, args):
        """Determine some docker tag-friendly version label for the containers to be built"""
        # For now use latest. Once we are in production we will switch to using semver using git describe
        return args.docker_tag

    def get_images_to_be_built(self, args):
        version = self.determine_version(args)

        containers = []
        for key, container in self.containers.items():
            if container.is_built():
                containers.append(container.patch_build(args, self.path, version))
        return containers

    def patch(self, args):
        for key, container in self.containers.items():
            container.patch(regcred=args.regcred)

    def lint(self):
        # Was there any failures?
        success = not any(
            map(lambda kv: not kv[1].lint(kv[0]), self.containers.items())
        )

        if not success:
            logger.error("Failed to validate %s" % self.relative_path())

        return success

    def has_containers(self):
        return len(self.containers.keys()) > 0

    def has_dynamic_challenges(self):
        return any([container.is_dynamic() for k, container in self.containers.items()])

    def has_static_challenges(self):
        return any(
            [(not container.is_dynamic()) for k, container in self.containers.items()]
        )

    def serialize(self, args):
        logger.debug(f"Serializing {self.relative_path()}")
        return {
            "relative_path": self.relative_path(),
            "challenge_id": self.challenge_id(),
            "has_containers": len(self.containers.keys()) > 0,
            "has_dynamic_containers": self.has_dynamic_challenges(),
            "has_static_containers": self.has_static_challenges(),
            "services": [
                self.containers[key].serialize(args.dns_host)
                for key in self.containers.keys()
            ],
        }


def enumerate_challenges(challengeroot: Path):
    logger.info("Enumerating challenges...")

    challenges = []

    for folder in challengeroot.glob("*/*"):
        if folder.is_file():
            logger.info(f"{folder} is file - skipping")
            continue
        if folder.parent.name in [".git", ".github"]:
            logger.info("Skipping %s" % folder)
            continue
        challenges.append(Challenge(folder))

    logger.info("Found %d challenges" % len(challenges))

    return challenges



from pathlib import Path

from dataclasses import dataclass

from different_wrapping.utils import docker_port_string_get_external_port

import re
import os

import logging

logger = logging.getLogger(__name__)

IMAGE_REPO = os.environ.get("IMAGE_REPO")


def normalize(text):
    return re.sub(r"-*$", "", re.sub(r"[^a-zA-Z0-9\-]", "-", text))


@dataclass
class DockerImage:
    dockerfile: Path
    context: Path
    url: str
    tags: list[str]

    def serialize(self):
        return {
            "dockerfile": str(self.dockerfile.absolute()),
            "context": str(self.context.absolute()),
            "url": self.url,
            "tags": self.tags,
        }


# Docker-compose linters
def service_name_requirement(container, svc_name):
    if re.match(r"^[a-zA-Z0-9\-]+$", svc_name) is None:
        return "Docker-compose service name must be in the form [a-zA-Z0-9\-]+ due to kompose renaming them otherwise."
    return None


def no_volumes(container, svc_name):
    if "volumes" in container.container_dict:
        return "No volumes are allowed. Build data into the docker container instead using COPY!"
    return None


def no_long_format_ports(container, svc_name):
    if "ports" in container.container_dict:
        for port in container.container_dict["ports"]:
            if isinstance(port, dict):
                return "Long-format port declarations are not allowed as they are not necessary"

    return None


docker_compose_linters = [service_name_requirement, no_volumes, no_long_format_ports]


class Container:
    def __init__(self, name: str, challenge, container_dict: dict):
        self.name = name
        self.challenge = challenge
        self.container_dict = container_dict

    def is_built(self):
        return "build" in self.container_dict

    def is_dynamic(self):
        return self.name.startswith("dynamic")

    def lint(self, name):
        success = True

        for linter in docker_compose_linters:
            status = linter(self, name)
            if status is not None:
                logger.error("Validation failed for %s: %s" % (name, status))
                success = False

        return success

    def patch_build(self, args, challenge_dir: Path, version_tag: str):
        """Patches the build part of a docker-compose service, turning it into an image URL"""
        # Handle different formats the dockerfile can be in
        build_info = self.container_dict["build"]
        dockerfile_name = (
            "Dockerfile" if type(build_info) is str else build_info["dockerfile"]
        )

        if type(build_info) is str:
            context = challenge_dir / Path(build_info)
        elif "context" not in build_info:
            context = challenge_dir
        else:
            context = challenge_dir / Path(build_info["context"])

        dockerfile = context / Path(dockerfile_name)

        image_name = normalize(dockerfile_name.lower().replace("dockerfile", ""))
        if len(image_name) > 0:
            image_url = "%s/%s/%s" % (
                args.docker_image_repo,
                normalize(challenge_dir.name),
                "f" + image_name,
            )
        else:
            image_url = "%s/%s" % (
                args.docker_image_repo,
                normalize(challenge_dir.name),
            )

        image_url = image_url.lower()

        logger.info(f"Determined docker image URL {image_url} for {challenge_dir.name}")
        del self.container_dict["build"]
        self.container_dict["image"] = "%s:%s" % (image_url, version_tag)

        return DockerImage(dockerfile, context, image_url, [version_tag])

    def patch(self, regcred=None):
        """Perform misc cleanups to the service"""
        if "ports" in self.container_dict:
            ports = self.container_dict["ports"]
            for idx, port in enumerate(ports):
                components = port.split(":")
                if len(components) == 1:
                    # No external port, nothing to do
                    continue
                elif len(components) == 2:
                    ports[idx] = components[1]
                elif len(components) == 3:
                    ports[idx] = components[2]

        if regcred is not None:
            self.container_dict["labels"]["kompose.image-pull-secret"] = regcred

    def has_label(self, target_label):
        logger.debug(self.container_dict)
        if "labels" not in self.container_dict:
            return False
        labels = self.container_dict["labels"]

        if isinstance(labels, dict):
            return target_label in labels
        elif isinstance(labels, list):
            for label in self.container_dict["labels"]:
                logger.debug(label.split("=")[0])
                if label.split("=")[0] == target_label:
                    return True
        else:
            raise RuntimeError("Lable is of invalid type")
        return False

    def get_label(self, target_label):
        logger.debug(self.container_dict)

        if "labels" not in self.container_dict:
            raise RuntimeError("Container does not have labels")

        labels = self.container_dict["labels"]

        # The label list can be both a dictionary or a list
        if isinstance(labels, dict):
            return labels[target_label]
        elif isinstance(labels, list):
            for label in self.container_dict["labels"]:
                split = label.split("=")
                if split[0] == target_label:
                    return split[1]

    def get_dns_name(self, dns_host):
        host = f"{self.name}-{self.challenge.safe_name()}.{dns_host}"
        if self.has_label("no.tghack.dns_challenge_name"):
            logger.debug(
                f"Challenge {self.name()} service {self.name} has custom DNS name - using it"
            )
            host = f"{self.get_label('no.tghack.dns_challenge_name')}.{dns_host}"

        return host

    def ports(self):
        return [
            docker_port_string_get_external_port(port)
            for port in self.container_dict["ports"]
        ]

    def serialize(self, dns_host):
        access_type = "none"

        access = None

        if self.has_label("no.tghack.http"):
            access = f"https://{self.get_dns_name(dns_host)}"
        elif self.has_label("no.tghack.tcp"):
            port = self.ports()[0]
            access = f"{self.get_dns_name(dns_host)}:{port}"

        if self.has_label("no.tghack.http"):
            access_type = "http"

        elif "ports" in self.container_dict and len(self.container_dict["ports"]) > 0:
            access_type = "tcpudp"

        base_dict = {
            "deploy_type": "dynamic" if self.is_dynamic() else "static",
            "access_type": access_type,
            "access": access,
            "random_password_env": self.get_label("no.tghack.autogenerated_password")
            if self.has_label("no.tghack.autogenerated_password")
            else None,
            "name": self.get_label("no.tghack.frontendName")
            if self.has_label("no.tghack.frontendName")
            else self.name,
            "svc_name": self.name,
        }

        return base_dict
